from io import StringIO
import os

from rayter.game_parser import GamesParser
from rayter.rater import Rater


def parse_game_file(path, slug):
    """
    Fetch game file and re-calculate ratings. If the game file does not exist, the data will be deleted.
    """
    with open(path, "r") as f:
        content = f.read()

    parser = GamesParser(StringIO(content), slug)
    matches = parser.parse_file()
    rater = Rater(matches)
    ratings = rater.rate_games(parser.score_type)

    if len(matches) > 0:
        game = {
            "slug": slug,
            "game_name": parser.game_name,
            "rating_history": {name:player.rating_history for name, player in rater.players.items()},
            "ratings": ratings,
            "count": len(matches),
            "score_type": parser.score_type
        }

        return game
    else:
        return None

def get_games(env):
    games = []

     # Create game list from game files
    for filename in os.listdir(env.games_path):
        if not filename.endswith(".txt"):
            continue
        if os.path.isdir(filename):
            continue

        # remove .txt from filename
        slug = filename[:-4]
        game = parse_game_file(os.path.join(env.games_path, filename), slug)

        if game is not None:
            games.append(game)

    return games



import numpy

def get_global_chart(games):
    global_placements = get_global_placements(games)
    return global_chart_placements(global_placements)

# Returns a dictionary of players and their placements in all games.
# Only games with at least 3 players are included.
# Only players with at least 3 matches in each of those games are included.
# The dictionary has the following structure:
# {
#   "player_name": [
#     {
#       "game_name": "A game name",
#       "normalized_placement": 0.5,
#       "game_count": 5
#     },
#     {
#       "game_name": "Another game name",
#       "normalized_placement": 0.25,
#       "game_count": 3
#     }
#   ]
# }
# The "game_count" value is the number of matches the player has played in the game.
# This is used to weight the placement in the average placement calculation.
# The "normalized_placement" value is the placement of the player in the game,
# normalized to a number between 0 and 1, inclusive.
# The lowest placement is 0, the highest is 1.
# The game name is included for debugging purposes and is currently not used.
def get_global_placements(games):
    global_placements = {}
    for game in games:
      players = game["ratings"]
      placement = 0
      # Only include games with at least 3 players
      if len(players) >= 3:
        for player_name, game_count, rating, delta in players:
            if player_name not in global_placements:
                global_placements[player_name] = []

            # Only include placements in games where the player has played at least 3 matches
            if (game_count >= 3):
                # normalize placement to a number between 0 and 1, inclusive
                # To make it inclusive, subtract 1 from the length of the players list.
                # If the players list contains only one player, this would lead to
                # division by zero. But that shouldn't happen, right...?
                normalized_placement = placement / float(len(players) - 1)
                global_placements[player_name].append({
                    "game_name": game["game_name"],
                    "normalized_placement": normalized_placement,
                    "game_count": game_count
                })
            placement = placement + 1

    return global_placements


# Returns an ordered list of players and their average placement in all games.
# Only players with at least 3 matches in each of their games are included.
# The list is ordered by the average placement, with the lowest average placement first.
# The list has the following structure:
# [
#   {
#     "player_name": "A player name",
#     "average": 0.5
#   },
#   {
#     "player_name": "Another player name",
#     "average": 0.25
#   }
# ]
# The "average" value is the average placement of the player in all games.
# This is calculated by taking the weighted average of the normalized placements
# in all games. The weight is the number of matches the player has played in the game.
# Weighted averages are used to give more weight to games where the player has played
# more matches.
def global_chart_placements(placements):
    average_placements = []
    for player_name, player_data in placements.items():
        # If player has played at least 3 games
        if len(player_data) >= 3:
            # Create a list of placements for all games
            player_placements = list(map(lambda placement: placement["normalized_placement"], player_data))
            # Create a matching list of the number of played matches for each game
            player_counts = list(map(lambda placement: placement["game_count"], player_data))

            average = numpy.average(player_placements, weights=player_counts)
            average_placements.append({
                "player_name": player_name,
                "average": average
              })

    average_placements.sort(key=lambda average_placement: average_placement["average"])

    return average_placements



import json
import os
import random
import requests
import shutil
import puremagic
import py_avataaars

def render_game_page(env, game, players):
    filename = os.path.join(env.output_path, game['slug'], 'index.html')
    env.render_template("game.html", {
        "name": game["game_name"],
        "ratings": game["ratings"],
        "game_name": game["game_name"],
        "slug": game["slug"],
        "players": players,
        "STATIC_URL": "../static/",
    }, output_path=filename)


def render_game_json(env, game):
    filename = os.path.join(env.output_path, game['slug'], 'game.json')
    with open(filename, 'w') as f:
        json.dump(game, f)

def render_player_page(env, player):
    filename = os.path.join(env.output_path, 'player', player['slug'], 'index.html')
    env.render_template("player.html", {
        "player": player,
        "STATIC_URL": "../../static/",
    }, output_path=filename)

# FIXME: This function is quite long. Should the main parts be broken out into
#        a separate file?
def render_player_image(env, player):
    dir_path = os.path.join(env.output_path, 'player', player['slug'])
    # Ensure the player folder exists.
    os.makedirs(dir_path, exist_ok=True)

    if player['imageUrl'] is None:
        # If there is no image, generate a random avatar using py_avataaars.

        # Create a seed from the player's name in order to make the avatar
        # deterministic.
        seed_string = "123" + player['name']
        seed = 0
        for c in seed_string:
            seed += (ord(c) % 100000)
        random.seed(seed)

        # Make facial hair less common.
        use_facial_hair = random.randint(0, 9) == 0

        # Create the random avatar.
        avatar = py_avataaars.PyAvataaar(
            style=py_avataaars.AvatarStyle.TRANSPARENT,
            skin_color=random.choice(list(py_avataaars.SkinColor)),
            hair_color=random.choice(list(py_avataaars.HairColor)),
            facial_hair_type=random.choice(list(py_avataaars.FacialHairType)) if use_facial_hair else py_avataaars.FacialHairType.DEFAULT,
            facial_hair_color=random.choice(list(py_avataaars.HairColor)),
            top_type=random.choice(list(py_avataaars.TopType)),
            hat_color=random.choice(list(py_avataaars.Color)),
            mouth_type=random.choice(list(py_avataaars.MouthType)),
            eye_type=random.choice(list(py_avataaars.EyesType)),
            eyebrow_type=random.choice(list(py_avataaars.EyebrowType)),
            nose_type=random.choice(list(py_avataaars.NoseType)),
            accessories_type=random.choice(list(py_avataaars.AccessoriesType)),
            clothe_type=random.choice(list(py_avataaars.ClotheType)),
            clothe_color=random.choice(list(py_avataaars.Color)),
            clothe_graphic_type=random.choice(list(py_avataaars.ClotheGraphicType)),
        )
        # And render it to the file.
        filename = 'image.png'
        file_path = os.path.join(dir_path, filename)

        print("Writing " + file_path)
        avatar.render_png_file(file_path)
    else:
        # Use the user's preferred image.
        image_url = player['imageUrl']

        # Create the correct extension.
        extension = image_url.split('.')[-1]
        # If the extension is too long, assume there was no extension.
        # Set it to a temporary extension and guess the type from the content
        # when we have downloaded the file.
        if len(extension) > 4:
            extension = 'temp'

        filename = 'image' + '.' + extension
        file_path = os.path.join(dir_path, filename)

        print("Getting " + image_url)

        # Some sites block requests without a user agent.
        headers = {'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/110.0.0.0 Safari/537.36'}

        # Download the image.
        response = requests.get(image_url, headers=headers)

        # Write the image to the file.
        print("Writing " + file_path)
        with open(file_path, 'wb') as fd:
            for chunk in response.iter_content(chunk_size=128):
                fd.write(chunk)

        # If the extension was temporary, guess the type from the content.
        if extension == 'temp':
            new_extension = puremagic.from_file(file_path)
            if new_extension is None:
                new_extension = '.png'

            # Rename the file to the correct extension.
            # Puremagic returns the extension with a dot so we don't need to add one.
            filename = 'image' + new_extension
            new_path = os.path.join(dir_path, filename)
            print("Mving 2 " + new_path)
            os.rename(file_path, new_path)

    # Update the player record with the new filename.
    player['image'] = filename

def render_index_page(env, games, players, global_chart):
    games.sort(key=lambda g: g["count"], reverse=True)
    filename = os.path.join(env.output_path, 'index.html')
    env.render_template("index.html", {
        "games": games,
        "players": players,
        "global_chart": global_chart,
        "log": [],
        "STATIC_URL": "static/",
    }, output_path=filename)



import hashlib
import urllib

def create_player_from_games(name, games):
    ratings = []

    for game in games:
        players = game["ratings"]
        placement = 0
        for player_name, _0, rating, _1 in players:
            if player_name == name:
                ratings.append(
                    (game["slug"], game["game_name"], rating, placement))
            placement += 1

    # Sort ratings by rating, best first. The actual rating is the third element in the tuple.
    ratings.sort(key=lambda rating: int(rating[2]), reverse=True)

    player = {
        "name": name,
        "slug": name,
        "ratings": ratings,
    }

    return player

# Creates an image url for a player. If the player has an email, use gravatar. Otherwise use a generated avatar.
def make_image_url(player):
    # Don't override explicit imageUrl
    if 'imageUrl' in player:
        return player['imageUrl']

    gravatar_size = 300

    if 'email' in player:
        email = player['email']
        hash = hashlib.md5(email.lower().encode("utf-8")).hexdigest()
        gravatar_url = "https://www.gravatar.com/avatar/" + hash + "?"
        gravatar_url += urllib.parse.urlencode({
            's': str(gravatar_size)
        })
        return gravatar_url
    else:
        return None


def get_all_players(games, env):
    players = {}
    players_metadata = env.players_metadata

    # Collect all players from all games
    for game in games:
        for player in game["ratings"]:
            name = player[0]
            if not name in players:
                players[name] = create_player_from_games(name, games)

    # Overwrite generic player values with player metadata from file
    if players_metadata != None:
        for player_name in players_metadata.keys():
            # If player exists in players, update it with all values in the player's metadata
            if player_name in players:
                player_metadata = players_metadata[player_name]
                players[player_name].update(player_metadata)

        # For every player, if they have no image url, try to create one from their metadata
        # FIXME: Should this be moved into create_player_from_games()?
        for player_name in players.keys():
            player = players[player_name]
            if not 'imageUrl' in player:
                player['imageUrl'] = make_image_url(player)

    return players



import json
import os

# try to import tomlib, if it fails, use tomli
try:
    import tomllib
except ImportError:
    import tomli as tomllib

from jinja2 import Environment, FileSystemLoader


def get_toml(path):
    with open(path, "rb") as f:
        return tomllib.load(f)

def get_json(path):
    with open(path, "rb") as f:
        return json.load(f)


class GeneratorEnvironment:
    def __init__(self, games_path, output_path, config_path=None, players_path=None):
        self.games_path = games_path
        self.output_path = output_path
        self.config_path = config_path
        self.players_path = players_path
        self.root_path = os.path.join(os.path.abspath(os.path.dirname(__file__)))

        self.config = {}
        if self.config_path:
            self.config = get_toml(self.config_path)
        elif os.path.exists("./config/rayter.toml"):
            self.config = get_toml("./config/rayter.toml")
        elif os.path.exists("./rayter.toml"):
            self.config = get_toml("./rayter.toml")

        if "site_name" not in self.config:
            self.config["site_name"] = "Rayter"
        if "site_description" not in self.config:
            self.config["site_description"] = "Find game ratings for our games here!"

        if self.players_path:
            self.players_metadata = get_json(self.players_path)
        elif os.path.exists("./config/players.json"):
            self.players_metadata = get_json("./config/players.json")
        elif os.path.exists("./players.json"):
            self.players_metadata = get_json("./players.json")
        else:
            self.players_metadata = {}

        self.jinja_env = Environment(loader=FileSystemLoader(os.path.join(self.root_path, "templates")))
        self.jinja_env.globals.update({
            "CONFIG": self.config,
        })

    def render_template(self, template_path, context, output_path=None):
        contents = self.jinja_env.get_template(template_path).render(context)
        if output_path:
            print('Writing', output_path)
            os.makedirs(os.path.dirname(output_path), exist_ok=True)
            with open(output_path, "w") as f:
                f.write(contents)
        return contents
